from typing import Any, List, Dict
from unittest import TestCase
from unittest.mock import Mock, patch, call

import click

from montecarlodata.utils import AwsClientWrapper

_SAMPLE_FILE_PATH = 'file'
_SAMPLE_BUCKET_NAME = 'bucket'
_SAMPLE_OBJECT_NAME = 'obj'
_SAMPLE_REGION = 'bar'
_SAMPLE_PROFILE_NAME = 'foo'
_SAMPLE_STACK_ID = 'arn:aws:stack'
_SAMPLE_STACK_LINK = 'link'
_SAMPLE_STACK_PARAMS = [{'foo': 'bar'}]
_SAMPLE_STACK_CAPABILITIES = 'CAPABILITY_IAM'
_SAMPLE_STACK_WAIT_CONFIG = {'Delay': 5, 'MaxAttempts': 720}
_SAMPLE_GATEWAY_ID = 'id'
_SAMPLE_GATEWAY_STAGE = 'stage'
_SAMPLE_STACK = {'Stacks': [{'Outputs': [0, 1, 2]}]}
_SAMPLE_TERM_PROTECT = False


class MockClient:
    # Arguments, classes and functions use AWS casing...

    @staticmethod
    def describe_stacks(StackName: str) -> Any:
        assert StackName == _SAMPLE_STACK_ID
        return _SAMPLE_STACK

    @staticmethod
    def upload_file(file_path: str, bucket_name: str, object_name: str):
        assert _SAMPLE_FILE_PATH == file_path
        assert _SAMPLE_BUCKET_NAME == bucket_name
        assert _SAMPLE_OBJECT_NAME == object_name

    @staticmethod
    def update_stack(StackName: str, TemplateURL: str, Parameters: List[Dict], Capabilities: List[str]):
        assert StackName == _SAMPLE_STACK_ID
        assert TemplateURL == _SAMPLE_STACK_LINK
        assert Parameters == _SAMPLE_STACK_PARAMS
        assert Capabilities == [_SAMPLE_STACK_CAPABILITIES]

    @staticmethod
    def create_stack(StackName: str, TemplateURL: str, Parameters: List[Dict], Capabilities: List[str], **kwargs):
        assert StackName == _SAMPLE_STACK_ID
        assert TemplateURL == _SAMPLE_STACK_LINK
        assert Parameters == _SAMPLE_STACK_PARAMS[0]
        assert Capabilities == [_SAMPLE_STACK_CAPABILITIES]
        return {'StackId': StackName}

    @staticmethod
    def create_deployment(restApiId: str, stageName: str):
        assert restApiId == _SAMPLE_GATEWAY_ID
        assert stageName == _SAMPLE_GATEWAY_STAGE

    class get_waiter:
        def __init__(self, waiter):
            self.wait = self._wait

        def _wait(self, StackName, WaiterConfig):
            assert StackName == _SAMPLE_STACK_ID
            assert WaiterConfig == _SAMPLE_STACK_WAIT_CONFIG


class AwsUtilTest(TestCase):
    def setUp(self) -> None:
        self._session_mock = Mock()

        self._service = self._setup()
        self._session_mock.reset_mock()

        self._session_mock.client.return_value = MockClient

    def test_setup(self):
        self._setup()

    def test_get_stack_outputs(self):
        self._service.get_stack_outputs(_SAMPLE_STACK_ID)
        self._session_mock.client.assert_called_once_with('cloudformation')

    def test_get_stack_parameters(self):
        self._service.get_stack_parameters(_SAMPLE_STACK_ID)
        self._session_mock.client.assert_called_once_with('cloudformation')

    @patch.object(AwsClientWrapper, 'get_stack_details')
    def test_upgrade_stack(self, mock_get_stack_details):
        mock_get_stack_details.return_value = {'Stacks': [{'StackStatus': 'UPDATE_COMPLETE'}]}
        self.assertTrue(self._service.upgrade_stack(_SAMPLE_STACK_ID, _SAMPLE_STACK_LINK, _SAMPLE_STACK_PARAMS))
        self._session_mock.client.assert_has_calls = [call('cloudformation') * 2]

    @patch.object(AwsClientWrapper, 'get_stack_details')
    def test_upgrade_stack_failed(self, mock_get_stack_details):
        mock_get_stack_details.return_value = {'Stacks': [{'StackStatus': 'UPDATE_FAILED'}]}
        self.assertFalse(self._service.upgrade_stack(_SAMPLE_STACK_ID, _SAMPLE_STACK_LINK, _SAMPLE_STACK_PARAMS))
        self._session_mock.client.assert_has_calls = [call('cloudformation') * 2]

    @patch.object(AwsClientWrapper, 'get_stack_details')
    def test_create_stack(self, mock_get_stack_details):
        mock_get_stack_details.return_value = {'Stacks': [{'StackStatus': 'CREATE_COMPLETE'}]}
        self.assertTrue(self._service.create_stack(_SAMPLE_STACK_ID, _SAMPLE_STACK_LINK, False, _SAMPLE_STACK_PARAMS[0]))
        self._session_mock.client.assert_has_calls = [call('cloudformation') * 2]

    @patch.object(AwsClientWrapper, 'get_stack_details')
    def test_create_stack_failed(self, mock_get_stack_details):
        mock_get_stack_details.return_value = {'Stacks': [{'StackStatus': 'CREATE_FAILED'}]}
        self.assertFalse(self._service.create_stack(_SAMPLE_STACK_ID, _SAMPLE_STACK_LINK, False, _SAMPLE_STACK_PARAMS[0]))
        self._session_mock.client.assert_has_calls = [call('cloudformation') * 2]

    def test_deploy_gateway(self):
        self._service.deploy_gateway(_SAMPLE_GATEWAY_ID, _SAMPLE_GATEWAY_STAGE)
        self._session_mock.client.assert_called_once_with('apigateway')

    def test_upload_file(self):
        self._service.upload_file(file_path=_SAMPLE_FILE_PATH,
                                  bucket_name=_SAMPLE_BUCKET_NAME,
                                  object_name=_SAMPLE_OBJECT_NAME)
        self._session_mock.client.assert_called_once_with('s3')

    def test_region_error(self):
        self._session_mock.client.side_effect = ValueError()
        with self.assertRaises(click.exceptions.Abort):
            self._setup()

    def _setup(self) -> AwsClientWrapper:
        client = AwsClientWrapper(_SAMPLE_PROFILE_NAME, _SAMPLE_REGION, session=self._session_mock)

        self._session_mock.client.assert_called_once_with('cloudformation')
        return client



import base64
import copy
from unittest import TestCase
from unittest.mock import Mock, patch, call

import click
from box import Box

from montecarlodata.common.data import MonolithResponse
from montecarlodata.common.user import UserService
from montecarlodata.integrations.onboarding.bi.reports import ReportsOnboardingService
from montecarlodata.queries.onboarding import TEST_TABLEAU_CRED_MUTATION, ADD_TABLEAU_CONNECTION_MUTATION, \
    TEST_LOOKER_METADATA_CRED_MUTATION, ADD_BI_CONNECTION_MUTATION, TEST_LOOKER_GIT_CLONE_CRED_MUTATION
from montecarlodata.utils import GqlWrapper, AwsClientWrapper
from tests.test_base_onboarding import _SAMPLE_BASE_OPTIONS
from tests.test_common_user import _SAMPLE_CONFIG

_SAMPLE_OPTIONS = {'foo': 'bar'}


class ReportOnboardingTest(TestCase):
    def setUp(self) -> None:
        self._user_service_mock = Mock(autospec=UserService)
        self._request_wrapper_mock = Mock(autospec=GqlWrapper)
        self._aws_wrapper_mock = Mock(autospec=AwsClientWrapper)

        self._service = ReportsOnboardingService(
            _SAMPLE_CONFIG,
            request_wrapper=self._request_wrapper_mock,
            aws_wrapper=self._aws_wrapper_mock,
            user_service=self._user_service_mock,
        )

    @patch('montecarlodata.integrations.onboarding.bi.reports.click')
    @patch('montecarlodata.integrations.onboarding.bi.reports.prompt_connection')
    def test_add_tableau(self, prompt_connection_mock, click_mock):
        self._request_wrapper_mock.make_request_v2.side_effect = [
            MonolithResponse(data=Box({'success': True})),
            MonolithResponse(data=Box({'tableauAccount': {'uuid': True}}))
        ]

        self._service.onboard_tableau(**_SAMPLE_OPTIONS)
        self._request_wrapper_mock.assert_has_calls(
            [
                call.make_request_v2(
                    query=TEST_TABLEAU_CRED_MUTATION,
                    operation='testTableauCredentials',
                    variables=_SAMPLE_OPTIONS
                ),
                call.make_request_v2(
                    query=ADD_TABLEAU_CONNECTION_MUTATION,
                    operation='addTableauAccount',
                    variables=_SAMPLE_OPTIONS
                )
            ]
        )
        prompt_connection_mock.assert_called_once_with(
            message='Please confirm you want to add this connection',
            skip_prompt=False
        )
        click_mock.echo.assert_has_calls(
            [
                call('Connection test was successful!'),
                call('Success! Added connection for tableau.')
            ]
        )

    @patch('montecarlodata.integrations.onboarding.bi.reports.click')
    @patch('montecarlodata.integrations.onboarding.bi.reports.prompt_connection')
    def test_add_tableau_with_validation_failure(self, prompt_connection_mock, click_mock):
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(data=Box({'success': False}))

        with self.assertRaises(click.exceptions.Abort):
            self._service.onboard_tableau(**_SAMPLE_OPTIONS)
        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=TEST_TABLEAU_CRED_MUTATION,
            operation='testTableauCredentials',
            variables=_SAMPLE_OPTIONS
        )
        prompt_connection_mock.assert_not_called()
        click_mock.echo.assert_not_called()

    @patch('montecarlodata.integrations.onboarding.bi.reports.click')
    @patch('montecarlodata.integrations.onboarding.bi.reports.prompt_connection')
    def test_add_tableau_with_validate_only(self, prompt_connection_mock, click_mock):
        options = copy.deepcopy(_SAMPLE_OPTIONS)
        options['validate_only'] = True
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(data=Box({'success': True}))

        self._service.onboard_tableau(**options)
        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=TEST_TABLEAU_CRED_MUTATION,
            operation='testTableauCredentials',
            variables=_SAMPLE_OPTIONS
        )
        prompt_connection_mock.assert_not_called()
        click_mock.echo.assert_has_calls(
            [
                call('Connection test was successful!'),
                call('Skipping adding the connection.')
            ]
        )

    @patch.object(ReportsOnboardingService, 'onboard')
    def test_looker_metadata_flow(self, onboard_mock):
        self._service.onboard_looker_metadata(**_SAMPLE_BASE_OPTIONS)
        onboard_mock.assert_called_once_with(
            validation_query=TEST_LOOKER_METADATA_CRED_MUTATION,
            connection_query=ADD_BI_CONNECTION_MUTATION,
            validation_response='testLookerCredentials',
            connection_response='addBiConnection',
            connection_type='looker',
            **_SAMPLE_BASE_OPTIONS
        )

    @patch.object(ReportsOnboardingService, 'onboard')
    @patch('montecarlodata.integrations.onboarding.bi.reports.read_as_base64')
    def test_looker_git_flow(self, read_as_base64_mock, onboard_mock):
        file_path, service_json = 'foo', 'bar'
        base64_key = base64.b64encode(service_json.encode('utf-8'))

        input_options = {'ssh_key': file_path, **_SAMPLE_BASE_OPTIONS}
        expected_options = {**{'ssh_key': base64_key.decode()}, **_SAMPLE_BASE_OPTIONS}

        read_as_base64_mock.return_value = base64_key

        self._service.onboard_looker_git(**input_options)
        read_as_base64_mock.assert_called_once_with(file_path)
        onboard_mock.assert_called_once_with(
            validation_query=TEST_LOOKER_GIT_CLONE_CRED_MUTATION,
            connection_query=ADD_BI_CONNECTION_MUTATION,
            validation_response='testLookerGitCloneCredentials',
            connection_response='addBiConnection',
            connection_type='looker-git-clone',
            **expected_options
        )



from typing import Optional

import click

from montecarlodata.common.common import read_as_json_string
from montecarlodata.config import Config
from montecarlodata.errors import complain_and_abort, manage_errors
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import EXPECTED_TOGGLE_EVENTS_GQL_RESPONSE_FIELD
from montecarlodata.queries.onboarding import TOGGLE_EVENT_MUTATION


class EventsOnboardingService(BaseOnboardingService):

    def __init__(self, config: Config, **kwargs):
        super().__init__(config, **kwargs)

    @manage_errors
    def toggle_event_configuration(self, **kwargs) -> Optional[bool]:
        """
        Toggle event configuration (effectively onboarding it if enable is set to true)
        """
        mapping_file = kwargs.pop('mapping_file', None)
        if mapping_file:
            kwargs['mapping'] = read_as_json_string(mapping_file)

        connection_type = kwargs.get('connection_type')
        if connection_type:
            kwargs['connection_type'] = connection_type.lower()

        response = self._request_wrapper.make_request_v2(
            query=TOGGLE_EVENT_MUTATION,
            operation=EXPECTED_TOGGLE_EVENTS_GQL_RESPONSE_FIELD,
            variables=kwargs
        )
        if response.data.success:
            click.echo(f"Success! {'Enabled' if kwargs['enable'] else 'Disabled'} events.")
            return True
        complain_and_abort(f'Failed to toggle events!')



from typing import Dict

from montecarlodata.common.data import OnboardingConfiguration
from montecarlodata.config import Config
from montecarlodata.errors import manage_errors, complain_and_abort
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import EXPECTED_DBT_CLOUD_RESPONSE_FIELD, \
    EXPECTED_ADD_CONNECTION_RESPONSE_FIELD, DBT_CLOUD_CONNECTION_TYPE
from montecarlodata.queries.onboarding import TEST_DBT_CLOUD_CRED_MUTATION, ADD_CONNECTION_MUTATION


class DbtCloudOnboardingService(BaseOnboardingService):
    def __init__(self, config: Config, **kwargs):
        super().__init__(config, **kwargs)

    @manage_errors
    def onboard_dbt_cloud(self, **kwargs) -> None:
        self.onboard(validation_query=TEST_DBT_CLOUD_CRED_MUTATION,
                     validation_response=EXPECTED_DBT_CLOUD_RESPONSE_FIELD,
                     connection_query=ADD_CONNECTION_MUTATION,
                     connection_response=EXPECTED_ADD_CONNECTION_RESPONSE_FIELD,
                     connection_type=DBT_CLOUD_CONNECTION_TYPE, **kwargs)

    def _disambiguate_warehouses(self, warehouse_type: str, onboarding_config: OnboardingConfiguration) -> Dict:
        """
        Determine which warehouse to associate with dbt cloud connection
        """
        if not self._user_service.warehouses:
            complain_and_abort('You must have at least one warehouse connection configured.')

        if onboarding_config.connection_id:
            # Return ID of warehouse containing the matching connection_id
            for warehouse in self._user_service.warehouses:
                for connection in warehouse['connections']:
                    if connection['uuid'] == onboarding_config.connection_id:
                        return {'dwId': warehouse['uuid']}
            complain_and_abort(f"Could not find connection ID = {onboarding_config.connection_id}"
                               "Hint: Use 'montecarlo integrations list' to find a connection ID of the desired warehouse")

        # If there is only a single warehouse
        if len(self._user_service.warehouses) == 1:
            return {'dwId': self._user_service.warehouses[0]['uuid']}

        complain_and_abort("Multiple warehouses detected. You must choose a warehouse connection using --connection-id")


from unittest import TestCase
from unittest.mock import Mock, patch

from box import Box
from pycarlo.core import Client

from montecarlodata.common.user import UserService
from montecarlodata.integrations.onboarding.data_lake.databricks_metastore import DatabricksMetastoreOnboardingService, \
    DEFAULT_GATEWAY_URL, DEFAULT_SECRET_NAME, DEFAULT_SECRET_SCOPE
from montecarlodata.queries.onboarding import TEST_DATABRICKS_CRED_MUTATION
from montecarlodata.utils import GqlWrapper, AwsClientWrapper
from tests.test_base_onboarding import _SAMPLE_BASE_OPTIONS
from tests.test_common_user import _SAMPLE_CONFIG


class DatabricksMetastoreOnboardingTest(TestCase):
    def setUp(self) -> None:
        self._mc_client_mock = Mock(autospec=Client)
        self._user_service_mock = Mock(autospec=UserService)
        self._request_wrapper_mock = Mock(autospec=GqlWrapper)
        self._aws_wrapper_mock = Mock(autospec=AwsClientWrapper)

        self._service = DatabricksMetastoreOnboardingService(
            _SAMPLE_CONFIG,
            mc_client=self._mc_client_mock,
            request_wrapper=self._request_wrapper_mock,
            aws_wrapper=self._aws_wrapper_mock,
            user_service=self._user_service_mock
        )

    @patch.object(DatabricksMetastoreOnboardingService, 'onboard')
    def test_databricks_metastore_flow(self, onboard_mock):
        options = {
            'databricks_workspace_url': 'databricks_workspace_url',
            'databricks_workspace_id': 'databricks_workspace_id',
            'databricks_cluster_id': 'databricks_cluster_id',
            'databricks_token': 'databricks_token',
            'databricks_secret_key': DEFAULT_SECRET_NAME,
            'databricks_secret_scope': DEFAULT_SECRET_SCOPE
        }
        job_info = dict(
            databricks_job_id='databricks_job_id',
            databricks_job_name='databricks_job_name',
            databricks_notebook_path='databricks_notebook_path'
        )

        self._mc_client_mock.return_value = Box(dict(
            create_databricks_notebook_job=dict(
                databricks=dict(
                    workspace_job_id=job_info['databricks_job_id'],
                    workspace_job_name=job_info['databricks_job_name'],
                    workspace_notebook_path=job_info['databricks_notebook_path']
                )
            )
        ))

        self._service.onboard_databricks_metastore(**options, **_SAMPLE_BASE_OPTIONS)

        expected_options = {**options, **_SAMPLE_BASE_OPTIONS}
        expected_job_limits = {
            **job_info,
            'integration_gateway': {
                'gateway_url': DEFAULT_GATEWAY_URL,
                'databricks_secret_key': DEFAULT_SECRET_NAME,
                'databricks_secret_scope': DEFAULT_SECRET_SCOPE
            }
        }

        onboard_mock.assert_called_once_with(
            validation_query=TEST_DATABRICKS_CRED_MUTATION,
            validation_response='testDatabricksCredentials',
            connection_type='databricks-metastore',
            job_limits=expected_job_limits,
            **expected_options
        )
        self.assertEqual(self._mc_client_mock.call_count, 2)



import json
import uuid
from typing import Dict
from unittest import TestCase
from unittest.mock import Mock, patch

from montecarlodata.common.data import MonolithResponse
from montecarlodata.common.user import UserService
from montecarlodata.integrations.onboarding.fields import EXPECTED_UPDATE_CREDENTIALS_RESPONSE_FIELD, \
    EXPECTED_REMOVE_CONNECTION_RESPONSE_FIELD, OPERATION_ERROR_VERBIAGE
from montecarlodata.integrations.onboarding.operations.connection_ops import ConnectionOperationsService
from montecarlodata.queries.onboarding import UPDATE_CREDENTIALS_MUTATION, REMOVE_CONNECTION_MUTATION, \
    TEST_EXISTING_CONNECTION_QUERY
from montecarlodata.utils import GqlWrapper, AwsClientWrapper
from tests.helpers import capture_function
from tests.test_common_user import _SAMPLE_CONFIG

_SAMPLE_UUID = uuid.uuid4()


class ConnectionOperationsTest(TestCase):
    def setUp(self) -> None:
        self._user_service_mock = Mock(autospec=UserService)
        self._request_wrapper_mock = Mock(autospec=GqlWrapper)
        self._aws_wrapper_mock = Mock(autospec=AwsClientWrapper)

        self._service = ConnectionOperationsService(
            _SAMPLE_CONFIG,
            request_wrapper=self._request_wrapper_mock,
            aws_wrapper=self._aws_wrapper_mock,
            user_service=self._user_service_mock
        )

    def test_update_credentials(self):
        changes = {'host': 'foo'}
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(data={'success': True})
        func = capture_function(self._service.update_credentials, dict(connection_id=_SAMPLE_UUID, changes=changes))

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=UPDATE_CREDENTIALS_MUTATION,
            operation=EXPECTED_UPDATE_CREDENTIALS_RESPONSE_FIELD,
            variables=dict(connection_id=str(_SAMPLE_UUID), changes=json.dumps(changes), should_validate=True, should_replace=False)
        )
        self.assertEqual(func.std_out.getvalue().strip(), f"Success! Updated '{_SAMPLE_UUID}'.")

    def test_remove_connection(self):
        self._test_remove_connection(params=dict(connection_id=_SAMPLE_UUID, no_prompt=True))

    @patch('montecarlodata.integrations.onboarding.operations.connection_ops.click.confirm')
    def test_remove_connection_with_prompt(self, confirm):
        confirm.return_value = True
        self._test_remove_connection(params=dict(connection_id=_SAMPLE_UUID, no_prompt=False))

    def _test_remove_connection(self, params: Dict):
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(data={'success': True})
        func = capture_function(self._service.remove_connection, params)

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=REMOVE_CONNECTION_MUTATION,
            operation=EXPECTED_REMOVE_CONNECTION_RESPONSE_FIELD,
            variables=dict(connection_id=str(_SAMPLE_UUID))
        )
        self.assertEqual(func.std_out.getvalue().strip(), f"Success! Removed '{_SAMPLE_UUID}'.")

    def test_echo_operation_status_with_errors(self):
        func = capture_function(
            function=self._service.echo_operation_status,
            params=dict(response=MonolithResponse(data={'success': False}), operation='foo')
        )  # Note - other tests capture the success state

        self.assertEqual(func.std_out.getvalue().strip(), OPERATION_ERROR_VERBIAGE)

    @patch('montecarlodata.integrations.onboarding.operations.connection_ops.click')
    def test_echo_test_existing(self, click_mock):
        connection_id = uuid.uuid4()
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(data={'success': True})
        self._service.echo_test_existing(connection_id=connection_id)
        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=TEST_EXISTING_CONNECTION_QUERY,
            operation='testExistingConnection',
            variables={'connection_id': connection_id}
        )
        click_mock.echo.assert_called_once_with('{\n    "success": true\n}')



from unittest import TestCase
from unittest.mock import Mock, patch

from montecarlodata.common.user import UserService
from montecarlodata.integrations.onboarding.data_lake.spark import SparkOnboardingService, \
    SPARK_BINARY_MODE_CONFIG_TYPE, SPARK_HTTP_MODE_CONFIG_TYPE, SPARK_DATABRICKS_CONFIG_TYPE
from montecarlodata.queries.onboarding import TEST_SPARK_BINARY_MODE_CRED_MUTATION, \
    TEST_SPARK_HTTP_MODE_CRED_MUTATION, TEST_SPARK_DATABRICKS_CRED_MUTATION
from montecarlodata.utils import GqlWrapper, AwsClientWrapper
from tests.test_base_onboarding import _SAMPLE_BASE_OPTIONS
from tests.test_common_user import _SAMPLE_CONFIG


class SparkOnboardingTest(TestCase):
    def setUp(self) -> None:
        self._user_service_mock = Mock(autospec=UserService)
        self._request_wrapper_mock = Mock(autospec=GqlWrapper)
        self._aws_wrapper_mock = Mock(autospec=AwsClientWrapper)

        self._service = SparkOnboardingService(
            _SAMPLE_CONFIG,
            request_wrapper=self._request_wrapper_mock,
            aws_wrapper=self._aws_wrapper_mock,
            user_service=self._user_service_mock
        )

    @patch.object(SparkOnboardingService, 'onboard')
    def test_spark_binary_mode_flow(self, onboard_mock):
        options = {'host': 'host', 'database': 'database', 'port': 10000, 'username': 'user', 'password': 'password'}

        self._service.onboard_spark(SPARK_BINARY_MODE_CONFIG_TYPE, **options, **_SAMPLE_BASE_OPTIONS)
        self._assert_onboard(TEST_SPARK_BINARY_MODE_CRED_MUTATION, options, onboard_mock)

    @patch.object(SparkOnboardingService, 'onboard')
    def test_spark_http_mode_flow(self, onboard_mock):
        options = {'url': 'url', 'username': 'user', 'password': 'password'}

        self._service.onboard_spark(SPARK_HTTP_MODE_CONFIG_TYPE, **options, **_SAMPLE_BASE_OPTIONS)
        self._assert_onboard(TEST_SPARK_HTTP_MODE_CRED_MUTATION, options, onboard_mock)

    @patch.object(SparkOnboardingService, 'onboard')
    def test_spark_databricks_flow(self, onboard_mock):
        options = {'workspace_url': 'workspace_url', 'workspace_id': 'workspace_id', 'cluster_id': 'cluster_id',
                   'token': 'token'}

        self._service.onboard_spark(SPARK_DATABRICKS_CONFIG_TYPE, **options, **_SAMPLE_BASE_OPTIONS)
        self._assert_onboard(TEST_SPARK_DATABRICKS_CRED_MUTATION, options, onboard_mock)

    @staticmethod
    def _assert_onboard(mutation, options, onboard_mock):
        expected_options = {**options, **_SAMPLE_BASE_OPTIONS, **{'connectionType': 'spark'}}

        onboard_mock.assert_called_once_with(
            validation_query=mutation,
            validation_response='testSparkCredentials',
            connection_type='spark',
            **expected_options
        )



import copy
import dataclasses
import json
from unittest import TestCase
from unittest.mock import Mock, patch, MagicMock, call, ANY

import click
from box import Box, BoxList

from montecarlodata.common.data import EventProperties, AWSArn, DcResourceProperties
from montecarlodata.common.resources import CloudResourceService
from montecarlodata.common.user import UserService
from montecarlodata.utils import AwsClientWrapper
from tests.helpers import capture_function
from tests.test_common_user import _SAMPLE_CONFIG

_SAMPLE_REGION = 'us-east-1'
_SAMPLE_BUCKET_NAME = 'bucket'
_SAMPLE_BASE_OPTIONS = {'foo': 'bar'}
_SAMPLE_DC_OUTPUT = [{'OutputKey': 'PrivateS3BucketArn', 'OutputValue': f'arn:aws:s3:::{_SAMPLE_BUCKET_NAME}'}]
_SAMPLE_CONNECTION_PATH = 'test'
_SAMPLE_ACCOUNT_ID = '1234'
_SAMPLE_EXTERNAL_ID = '5678'
_SAMPLE_TIME = 'ts'
_SAMPLE_POLICY_FILE = {'hello': 'world'}
_SAMPLE_TRUST_POLICY = {
    'Version': '2012-10-17',
    'Statement': [
        {
            'Effect': 'Allow',
            'Principal': {
                'AWS': f'arn:aws:iam::{_SAMPLE_ACCOUNT_ID}:root'
            },
            'Action': 'sts:AssumeRole',
            'Condition': {
                'StringEquals': {
                    'sts:ExternalId': _SAMPLE_EXTERNAL_ID
                }
            }
        }
    ]
}
_SAMPLE_EMR_NAME = 'name'
_SAMPLE_EMR_ID = 'cluster_id'
_SAMPLE_EMR_ID_2 = 'cluster_id_2'
_SAMPLE_EMR_STATE = 'WAITING'
_SAMPLE_EMR_STATUS = {'State': _SAMPLE_EMR_STATE}
_SAMPLE_EMR_LOG_URI = 's3://bucket/path'
_SAMPLE_EMR_LOG_URI_2 = 's3://bucket_2/path'

_SAMPLE_EVENT_PAYLOAD = {
    'bucket_name': 'mcd-test-events',
    'collector_aws_profile': 'dev',
    'prefix': None,
    'suffix': None,
    'topic_arn': None,
    'bucket_aws_profile': None,
    'event_type': 'metadata'
}
_SAMPLE_COLLECTOR = {
    'uuid': 'e34eaf86-092d-47ac-97c3-cfa8b6a6e7fe',
    'stackArn': 'arn:aws:cloudformation:us-east-1:123456789:stack/loki/99',
    'active': True,
    'customerAwsAccountId': '123456789',
    'templateVersion': '42',
    'codeVersion': '2042',
    'lastUpdated': None
}

_SAMPLE_DC_CLIENT = Mock()
_SAMPLE_CUST_RESOURCES_CLIENT = Mock()
_SAMPLE_EVENT_DETAILS = EventProperties(
    event_type='metadata',
    bucket_name='mcd-test-events',
    event_queue_arn='arn:aws:sqs:us-east-1:123456789:loki-MetadataEventQueue-42',
    collection_region='us-east-1',
    cust_resources_region='us-east-1',
    collection_account_id='123456789',
    cust_account_id='123456789',
    collection_client=_SAMPLE_DC_CLIENT,
    cust_resources_client=_SAMPLE_CUST_RESOURCES_CLIENT
)


class CloudResourcesTest(TestCase):
    def setUp(self) -> None:
        self._user_service_mock = Mock(autospec=UserService)
        self._aws_wrapper_mock = Mock(autospec=AwsClientWrapper)

        self._service = CloudResourceService(
            _SAMPLE_CONFIG,
            aws_wrapper=self._aws_wrapper_mock,
            user_service=self._user_service_mock
        )

    def tearDown(self) -> None:
        _SAMPLE_DC_CLIENT.reset_mock()
        _SAMPLE_CUST_RESOURCES_CLIENT.reset_mock()

    @patch.object(CloudResourceService, '_read_json')
    def test_create_role_with_invalid_policy(self, mock_read):
        mock_read.side_effect = json.decoder.JSONDecodeError
        self._user_service_mock.active_collector = {'customerAwsAccountId': _SAMPLE_ACCOUNT_ID}

        with self.assertRaises(click.exceptions.Abort):
            self._service.create_role(path_to_policy_doc='invalid')

    @patch.object(CloudResourceService, '_read_json')
    def test_create_role_with_missing_values(self, mock_read):
        mock_read.return_value = _SAMPLE_POLICY_FILE
        self._user_service_mock.active_collector = {}
        with self.assertRaises(click.exceptions.Abort):
            self._service.create_role(path_to_policy_doc='foo')

    @patch.object(CloudResourceService, '_read_json')
    @patch.object(CloudResourceService, '_generate_random_token')
    @patch('time.time', MagicMock(return_value=_SAMPLE_TIME))
    def test_create_role(self, mock_token, mock_read):
        path_to_policy_doc = 'foo'
        mock_read.return_value = _SAMPLE_POLICY_FILE
        mock_token.return_value = _SAMPLE_EXTERNAL_ID
        self._user_service_mock.active_collector = {'customerAwsAccountId': _SAMPLE_ACCOUNT_ID}

        self._service.create_role(path_to_policy_doc=path_to_policy_doc)

        mock_read.assert_called_once_with(path_to_policy_doc)
        mock_token.assert_called_once_with()

        self._aws_wrapper_mock.create_role.assert_called_once_with(
            role_name=f'monte-carlo-integration-role-{_SAMPLE_TIME}',
            trust_policy=json.dumps(_SAMPLE_TRUST_POLICY),
            tags=self._service._MCD_TAGS
        )

        self._aws_wrapper_mock.attach_inline_policy.assert_called_once_with(
            role_name=f'monte-carlo-integration-role-{_SAMPLE_TIME}',
            policy_name=f'monte-carlo-integration-cli-policy-{_SAMPLE_TIME}',
            policy_doc=json.dumps(_SAMPLE_POLICY_FILE)
        )

    def test_generate_policy(self):
        policy = self._service._generate_trust_policy(account_id=_SAMPLE_ACCOUNT_ID, external_id=_SAMPLE_EXTERNAL_ID)

        self.assertEqual(policy, json.dumps(_SAMPLE_TRUST_POLICY))

    def test_list_emr_clusters(self):
        self._aws_wrapper_mock.get_emr_cluster_details.return_value = [{
            'Name': _SAMPLE_EMR_NAME,
            'Id': _SAMPLE_EMR_ID,
            'Status': _SAMPLE_EMR_STATUS,
            'LogUri': _SAMPLE_EMR_LOG_URI
        }]

        std_out = capture_function(self._service.list_emr_clusters, {'table_format': 'plain'}).std_out

        expected_output = ('Name    Id          State    LogUri\n'
                           f'{_SAMPLE_EMR_NAME}    {_SAMPLE_EMR_ID}  {_SAMPLE_EMR_STATE}  {_SAMPLE_EMR_LOG_URI}')
        self.assertEqual(std_out.getvalue().strip(), expected_output)

    def test_list_emr_clusters_no_wait(self):
        self._aws_wrapper_mock.get_emr_cluster_details.return_value = [{
            'Name': _SAMPLE_EMR_NAME,
            'Id': _SAMPLE_EMR_ID,
            'Status': _SAMPLE_EMR_STATUS,
            'LogUri': _SAMPLE_EMR_LOG_URI
        }]

        std_out = capture_function(self._service.list_emr_clusters, {'no_grid': True}).std_out
        expected_output = '  '.join([_SAMPLE_EMR_NAME, _SAMPLE_EMR_ID, _SAMPLE_EMR_STATE, _SAMPLE_EMR_LOG_URI])
        self.assertEqual(std_out.getvalue().strip().strip(), expected_output)

    def test_list_emr_clusters_only_log_locations(self):
        self._aws_wrapper_mock.get_emr_cluster_details.return_value = [
            {
                'Name': _SAMPLE_EMR_NAME,
                'Id': _SAMPLE_EMR_ID,
                'Status': _SAMPLE_EMR_STATUS,
                'LogUri': _SAMPLE_EMR_LOG_URI
            },
            {
                'Name': _SAMPLE_EMR_NAME,
                'Id': _SAMPLE_EMR_ID_2,
                'Status': _SAMPLE_EMR_STATUS,
                'LogUri': _SAMPLE_EMR_LOG_URI_2
            }
        ]

        std_out = capture_function(self._service.list_emr_clusters, {'only_log_locations': True}).std_out
        self.assertEqual(set(std_out.getvalue().strip().split('\n')), {_SAMPLE_EMR_LOG_URI, _SAMPLE_EMR_LOG_URI_2})

    def test_list_emr_clusters_no_clusters(self):
        self._aws_wrapper_mock.get_emr_cluster_details.return_value = []
        self._user_service_mock.active_collector = {'customerAwsAccountId': _SAMPLE_ACCOUNT_ID}

        std_out = capture_function(self._service.list_emr_clusters, {'table_format': 'plain'}).std_out
        self.assertEqual(std_out.getvalue().strip(), 'Name    Id    State    LogUri')

    @patch('montecarlodata.common.resources.click')
    @patch.object(CloudResourceService, '_initialize_events')
    @patch.object(CloudResourceService, '_create_or_update_event_topic')
    @patch.object(CloudResourceService, '_update_event_q')
    @patch.object(CloudResourceService, '_create_topic_subscription')
    @patch.object(CloudResourceService, '_create_event_notification')
    def test_add_events_flow(self, mock_notification, mock_topic_sub, mock_q, mock_topic, mock_setup, mock_click):
        self._user_service_mock.get_collector.return_value = Box(_SAMPLE_COLLECTOR)
        mock_setup.return_value = _SAMPLE_EVENT_DETAILS

        self._service.add_events(**_SAMPLE_EVENT_PAYLOAD)

        self._user_service_mock.get_collector.assert_called_once_with(dc_id=None)

        mock_setup.assert_called_once_with(collector_props=_SAMPLE_COLLECTOR, **_SAMPLE_EVENT_PAYLOAD)
        mock_topic.assert_called_once_with(event_details=_SAMPLE_EVENT_DETAILS, skip_prompts=False)
        mock_q.assert_called_once_with(event_details=_SAMPLE_EVENT_DETAILS, skip_prompts=False)
        mock_topic_sub.assert_called_once_with(event_details=_SAMPLE_EVENT_DETAILS, skip_prompts=False)
        mock_notification.assert_called_once_with(event_details=_SAMPLE_EVENT_DETAILS, skip_prompts=False)

        mock_click.assert_has_calls(
            [
                call.echo('[Gathering collector and resource details]'),
                call.echo('[Creating or updating topic properties]'),
                call.echo('[Updating event queue properties]'),
                call.echo('[Creating topic subscription]'),
                call.echo('[Creating event notification]'),
                call.echo('\nProcess complete. Have a nice day!')
            ]
        )

    def test_add_events_with_an_inactive_collector(self):
        dc_id = 'thor'
        inactive_collector = Box(copy.deepcopy(_SAMPLE_COLLECTOR))
        inactive_collector.active = False

        self._user_service_mock.get_collector.return_value = inactive_collector
        with self.assertRaises(click.exceptions.Abort):
            self._service.add_events(**_SAMPLE_EVENT_PAYLOAD, **{'dc_id': dc_id})
        self._user_service_mock.get_collector.assert_called_once_with(dc_id=dc_id)

    @patch('montecarlodata.common.resources.AwsClientWrapper')
    @patch.object(CloudResourceService, '_get_relevant_q')
    def test_initialize_events(self, mock_get_q, aws_wrapper):
        derived_bucket_location, dc_outputs, q_arn = 'thor', {'foo': 'bar'}, 'arn:aws:sqs:us-east-1:12345:odin'
        aws_wrapper().get_bucket_location.return_value = derived_bucket_location
        aws_wrapper().get_stack_outputs.return_value = dc_outputs
        aws_wrapper().get_caller_identity.return_value = AWSArn(q_arn).account

        mock_get_q.return_value = q_arn

        self.assertEqual(
            self._service._initialize_events(collector_props=Box(_SAMPLE_COLLECTOR), **_SAMPLE_EVENT_PAYLOAD),
            EventProperties(
                event_type=_SAMPLE_EVENT_PAYLOAD['event_type'],
                bucket_name=_SAMPLE_EVENT_PAYLOAD['bucket_name'],
                event_queue_arn=q_arn,
                collection_region=AWSArn(q_arn).region,
                cust_resources_region=derived_bucket_location,
                collection_account_id=AWSArn(_SAMPLE_COLLECTOR['stackArn']).account,
                cust_account_id=AWSArn(q_arn).account,
                collection_client=ANY,
                cust_resources_client=ANY
            )
        )

        aws_wrapper.assert_has_calls(
            [
                call(profile_name=_SAMPLE_CONFIG.aws_profile, region_name=_SAMPLE_CONFIG.aws_region),
                call().get_bucket_location(bucket_name='mcd-test-events'),
                call(profile_name=_SAMPLE_EVENT_PAYLOAD['collector_aws_profile'],
                     region_name=_SAMPLE_CONFIG.aws_region),
                call(profile_name=_SAMPLE_CONFIG.aws_profile, region_name=derived_bucket_location),
                call().get_stack_outputs(stack_id=_SAMPLE_COLLECTOR['stackArn']),
                call().get_caller_identity()
            ]
        )
        mock_get_q.assert_called_once_with('metadata', dc_outputs)

    def test_get_relevant_q(self):
        m_out, q_out = 'loki', 'thor'
        sample_outputs = [
            {'OutputKey': 'MetadataEventQueue', 'OutputValue': m_out, 'Description': 'Metadata event queue ARN'},
            {'OutputKey': 'PublicIP', 'OutputValue': '127.0.0.1', 'Description': 'Does stuff'},
            {'OutputKey': 'QueryLogEventQueue', 'OutputValue': q_out, 'Description': 'Query log event queue ARN'}
        ]

        self.assertEqual(self._service._get_relevant_q(event_type='metadata', outputs=BoxList(sample_outputs)), m_out)
        self.assertEqual(self._service._get_relevant_q(event_type='query-logs', outputs=BoxList(sample_outputs)), q_out)
        with self.assertRaises(click.exceptions.Abort):
            self._service._get_relevant_q(event_type='invalid', outputs=BoxList(sample_outputs))  # bad event type
        with self.assertRaises(click.exceptions.Abort):
            self._service._get_relevant_q(event_type='metadata', outputs=Box({}))  # missing event config

    @patch('montecarlodata.common.resources.prompt_connection')
    @patch.object(CloudResourceService, '_extend_topic_policy')
    @patch.object(CloudResourceService, '_is_preexisting_topic_policy')
    def test_create_or_update_event_topic(self, mock_is_preexisting, mock_extend_policy, mock_prompt):
        attributes = {'foo': 'bar'}
        topic_name = 'monte-carlo-data-us-east-1-metadata-events-topic'
        topic_arn = 'arn:aws:sns:us-east-1:12345:monte-carlo-data-us-east-1-metadata-events-topic'
        _SAMPLE_CUST_RESOURCES_CLIENT.create_sns_topic.return_value = topic_arn
        _SAMPLE_CUST_RESOURCES_CLIENT.get_topic_attributes.return_value = attributes
        mock_is_preexisting.return_value = False
        mock_extend_policy.return_value = attributes

        self._service._create_or_update_event_topic(event_details=_SAMPLE_EVENT_DETAILS, skip_prompts=True)
        mock_is_preexisting.assert_has_calls(
            [
                call(
                    topic_policy_id='monte-carlo-data-metadata-mcd-test-events-ID',
                    topic_attributes={'foo': 'bar'}
                ),
                call(
                    topic_policy_id='monte-carlo-data-metadata-loki-MetadataEventQueue-42-ID',
                    topic_attributes={'foo': 'bar'}
                )
            ]
        )
        mock_extend_policy.assert_called_once_with(
            topic_attributes={'foo': 'bar'},
            new_statements=[
                {'Sid': 'monte-carlo-data-metadata-mcd-test-events-ID', 'Effect': 'Allow', 'Principal': {'AWS': '*'},
                 'Action': 'SNS:Publish',
                 'Resource': 'arn:aws:sns:us-east-1:12345:monte-carlo-data-us-east-1-metadata-events-topic',
                 'Condition': {'StringEquals': {'aws:SourceArn': 'arn:aws:s3:::mcd-test-events'}}},
                {'Sid': 'monte-carlo-data-metadata-loki-MetadataEventQueue-42-ID', 'Effect': 'Allow',
                 'Principal': {'AWS': '123456789'}, 'Action': 'sns:Subscribe',
                 'Resource': 'arn:aws:sns:us-east-1:12345:monte-carlo-data-us-east-1-metadata-events-topic'}]
        )
        _SAMPLE_CUST_RESOURCES_CLIENT.set_topic_attributes.assert_called_once_with(
            arn=topic_arn,
            name='Policy',
            value=attributes
        )
        _SAMPLE_CUST_RESOURCES_CLIENT.get_topic_attributes.assert_called_once_with(arn=topic_arn)
        _SAMPLE_CUST_RESOURCES_CLIENT.create_sns_topic.assert_called_once_with(name=topic_name)
        mock_prompt.assert_has_calls(
            [
                call(
                    message=f"Create topic '{topic_name}' in 'us-east-1' (123456789)? "
                            "This is an idempotent operation.",
                    skip_prompt=True
                ),
                call(message='Please confirm', skip_prompt=True)
            ]
        )

    def test_create_or_update_event_topic_invalid(self):
        events = dataclasses.replace(_SAMPLE_EVENT_DETAILS)

        with self.assertRaises(click.exceptions.Abort):
            events.topic_arn = 'foo'
            self._service._create_or_update_event_topic(event_details=events, skip_prompts=True)  # invalid ARN

        with self.assertRaises(click.exceptions.Abort):
            events.topic_arn = 'arn:aws:sns:us-east-2:12345:monte-carlo-data-us-east-1-metadata-events-topic'
            self._service._create_or_update_event_topic(event_details=events, skip_prompts=True)  # region mismatch

    def test_preexisting_topic_policy(self):
        sid, topic_attributes = 'foo', Box({'Attributes': {'Policy': '{"Statement": []}'}})
        self.assertFalse(self._service._is_preexisting_topic_policy(sid, topic_attributes))

        topic_attributes = Box({'Attributes': {'Policy': '{"Statement": [{"Sid": "foo"}]}'}})
        self.assertTrue(self._service._is_preexisting_topic_policy(sid, topic_attributes))

    def test_extend_topic_policy(self):
        topic_attributes = Box({'Attributes': {'Policy': '{"Statement": [{"Sid": "foo"}]}'}})
        new_statements = [{'x': 'y'}]
        self.assertEqual(
            self._service._extend_topic_policy(topic_attributes, new_statements),
            '{"Statement": [{"Sid": "foo"}, {"x": "y"}]}'
        )

    @patch('montecarlodata.common.resources.prompt_connection')
    def test_update_event_q(self, mock_prompt):
        name = 'loki-MetadataEventQueue-42'
        _SAMPLE_DC_CLIENT.get_q_attributes.return_value = None

        self._service._update_event_q(event_details=_SAMPLE_EVENT_DETAILS, skip_prompts=True)

        _SAMPLE_DC_CLIENT.set_q_attributes.assert_called_once_with(
            name=name,
            attributes={
                'Policy': '{"Version": "2008-10-17", "Statement": [{"Sid": "__owner", "Effect": "Allow", "Principal": '
                          '{"AWS": "arn:aws:iam::123456789:root"}, "Action": "SQS:*", "Resource": '
                          '"arn:aws:sqs:us-east-1:123456789:loki-MetadataEventQueue-42"}, '
                          '{"Sid": "__sender", "Effect": "Allow", "Principal": {"AWS": "*"}, "Action": '
                          '"SQS:SendMessage", "Resource": "arn:aws:sqs:us-east-1:123456789:loki-MetadataEventQueue-42",'
                          ' "Condition": {"ArnLike": {"aws:SourceArn": '
                          '["arn:aws:sns:us-east-1:12345:monte-carlo-data-us-east-1-metadata-events-topic"]}}}]}'
            }
        )
        _SAMPLE_DC_CLIENT.get_q_attributes.assert_called_once_with(name=name, attributes=['Policy'])
        mock_prompt.assert_called_once_with(message='Please confirm', skip_prompt=True)

    @patch('montecarlodata.common.resources.prompt_connection')
    @patch.object(CloudResourceService, '_extend_q_policy')
    def test_update_event_q_with_existing(self, mock_extend, mock_prompt):
        name, mock_extend.return_value = 'loki-MetadataEventQueue-42', {'foo': 'bar'}

        _SAMPLE_DC_CLIENT.get_q_attributes.return_value = Box({'Attributes': {"Policy": '{}'}})
        self._service._update_event_q(event_details=_SAMPLE_EVENT_DETAILS, skip_prompts=True)

        _SAMPLE_DC_CLIENT.get_q_attributes.assert_called_once_with(name=name, attributes=['Policy'])
        _SAMPLE_DC_CLIENT.set_q_attributes.assert_called_once_with(name=name, attributes={'Policy': '{"foo": "bar"}'})
        mock_prompt.assert_called_once_with(message='Please confirm', skip_prompt=True)

    @patch('montecarlodata.common.resources.prompt_connection')
    @patch.object(CloudResourceService, '_extend_q_policy')
    def test_update_event_q_with_no_delta(self, mock_extend, mock_prompt):
        name, mock_extend.return_value = 'loki-MetadataEventQueue-42', {'foo': 'bar'}

        _SAMPLE_DC_CLIENT.get_q_attributes.return_value = Box({'Attributes': {"Policy": '{"foo": "bar"}'}})
        self._service._update_event_q(event_details=_SAMPLE_EVENT_DETAILS, skip_prompts=True)

        _SAMPLE_DC_CLIENT.get_q_attributes.assert_called_once_with(name=name, attributes=['Policy'])
        _SAMPLE_DC_CLIENT.set_q_attributes.assert_not_called()
        mock_prompt.assert_not_called()

    def test_extend_q_policy(self):
        events = dataclasses.replace(_SAMPLE_EVENT_DETAILS)
        events.topic_arn = 'bar'

        # Existing is string
        existing_policy = Box({'Statement': [{'Sid': '__sender', 'Condition': {'ArnLike': {'aws:SourceArn': 'foo'}}}]})
        self.assertEqual(
            self._service._extend_q_policy(existing_policy=existing_policy, event_details=events),
            {'Statement': [{'Sid': '__sender', 'Condition': {'ArnLike': {'aws:SourceArn': ['foo', 'bar']}}}]}
        )

        # Existing is list
        existing_policy = Box({'Statement': [{'Sid': '__sender', 'Condition': {'ArnLike': {'aws:SourceArn': ['foo', 'qux']}}}]})
        self.assertEqual(
            self._service._extend_q_policy(existing_policy=existing_policy, event_details=events),
            {'Statement': [{'Sid': '__sender', 'Condition': {'ArnLike': {'aws:SourceArn': ['foo', 'qux', 'bar']}}}]}
        )

        # Existing has duplicate
        existing_policy = Box({'Statement': [{'Sid': '__sender', 'Condition': {'ArnLike': {'aws:SourceArn': 'bar'}}}]})
        self.assertEqual(
            self._service._extend_q_policy(existing_policy=existing_policy, event_details=events),
            {'Statement': [{'Sid': '__sender', 'Condition': {'ArnLike': {'aws:SourceArn': 'bar'}}}]}
        )
        existing_policy = Box({'Statement': [{'Sid': '__sender', 'Condition': {'ArnLike': {'aws:SourceArn': ['foo', 'bar']}}}]})
        self.assertEqual(
            self._service._extend_q_policy(existing_policy=existing_policy, event_details=events),
            {'Statement': [{'Sid': '__sender', 'Condition': {'ArnLike': {'aws:SourceArn': ['foo', 'bar']}}}]}
        )

        # Incompatible format
        with self.assertRaises(click.exceptions.Abort):
            self._service._extend_q_policy(existing_policy=Box({'Statement': [{'Sid': 'unk'}]}), event_details=events)

    @patch('montecarlodata.common.resources.prompt_connection')
    @patch.object(CloudResourceService, '_merge_event_notifications')
    def test_create_event_notification(self, mock_merge_events, mock_prompt):
        events = dataclasses.replace(_SAMPLE_EVENT_DETAILS)
        events.topic_arn = 'arn:aws:sns:us-east-1:12345:monte-carlo-data-us-east-1-metadata-events-topic'
        events.prefix = 'lake'

        new_config, existing_config = {'foo': 'bar'}, {'bar': 'foo'}
        _SAMPLE_CUST_RESOURCES_CLIENT.get_bucket_event_config.return_value = existing_config
        mock_merge_events.return_value = new_config
        self._service._create_event_notification(event_details=events, skip_prompts=True)

        mock_merge_events.assert_called_once_with(
            setting_id='monte-carlo-data-us-east-1-metadata-events-topic-lake-notification',
            existing_config=existing_config,
            event_details=events
        )
        _SAMPLE_CUST_RESOURCES_CLIENT.get_bucket_event_config.assert_called_once_with(name=events.bucket_name)
        _SAMPLE_CUST_RESOURCES_CLIENT.set_bucket_event_config.assert_called_once_with(
            name=events.bucket_name,
            notification_config=new_config
        )
        mock_prompt.assert_called_once_with(message='Please confirm', skip_prompt=True)

    @patch('montecarlodata.common.resources.prompt_connection')
    @patch.object(CloudResourceService, '_merge_event_notifications')
    def test_create_event_notification_with_no_delta(self, mock_merge_events, mock_prompt):
        events = dataclasses.replace(_SAMPLE_EVENT_DETAILS)
        events.topic_arn = 'arn:aws:sns:us-east-1:12345:monte-carlo-data-us-east-1-metadata-events-topic'

        new_config, existing_config = {'foo': 'bar'}, {'foo': 'bar'}
        _SAMPLE_CUST_RESOURCES_CLIENT.get_bucket_event_config.return_value = existing_config
        mock_merge_events.return_value = new_config
        self._service._create_event_notification(event_details=events, skip_prompts=True)

        mock_merge_events.assert_called_once_with(
            setting_id='monte-carlo-data-us-east-1-metadata-events-topic-notification',
            existing_config=existing_config,
            event_details=events
        )
        _SAMPLE_CUST_RESOURCES_CLIENT.get_bucket_event_config.assert_called_once_with(name=events.bucket_name)
        _SAMPLE_CUST_RESOURCES_CLIENT.set_bucket_event_config.assert_not_called()
        mock_prompt.assert_not_called()

    def test_merge_event_notifications(self):
        setting_id = 'foo'
        expected_config = {
            'Id': setting_id,
            'TopicArn': 'arn:aws:sns:us-east-1:12345:monte-carlo-data-us-east-1-metadata-events-topic',
            'Events': ['s3:ObjectCreated:*', 's3:ObjectRemoved:*'],
            'Filter': {'Key': {'FilterRules': [{'Name': 'Prefix', 'Value': ''}, {'Name': 'Suffix', 'Value': ''}]}}
        }

        # ID already exists
        existing_setting = {'TopicConfigurations': [{'Id': setting_id}]}
        self.assertEqual(
            self._service._merge_event_notifications(setting_id, existing_setting, _SAMPLE_EVENT_DETAILS),
            existing_setting
        )

        # Topic already exits
        existing_setting = {'TopicConfigurations': [{'Id': 'bar'}], 'QueueConfigurations': [{'foo': 'bar'}]}
        self.assertEqual(
            self._service._merge_event_notifications(setting_id, existing_setting, _SAMPLE_EVENT_DETAILS),
            {'TopicConfigurations': [{'Id': 'bar'}, expected_config], 'QueueConfigurations': [{'foo': 'bar'}]}
        )

        # No topic config
        self.assertEqual(
            self._service._merge_event_notifications(setting_id, {}, _SAMPLE_EVENT_DETAILS),
            {'TopicConfigurations': [expected_config]}
        )
        existing_setting = {'QueueConfigurations': [{'foo': 'bar'}]}
        self.assertEqual(
            self._service._merge_event_notifications(setting_id, existing_setting, _SAMPLE_EVENT_DETAILS),
            {'TopicConfigurations': [expected_config], 'QueueConfigurations': [{'foo': 'bar'}]}
        )

    @patch('montecarlodata.common.resources.prompt_connection')
    def test_create_topic_subscription(self, mock_prompt):
        _SAMPLE_CUST_RESOURCES_CLIENT.list_topic_subscriptions.return_value = None
        self._service._create_topic_subscription(event_details=_SAMPLE_EVENT_DETAILS, skip_prompts=True)
        _SAMPLE_CUST_RESOURCES_CLIENT.list_topic_subscriptions.assert_called_once_with(
            arn=_SAMPLE_EVENT_DETAILS.topic_arn
        )
        _SAMPLE_CUST_RESOURCES_CLIENT.subscribe_to_topic.assert_called_once_with(
            arn=_SAMPLE_EVENT_DETAILS.topic_arn,
            endpoint=_SAMPLE_EVENT_DETAILS.event_queue_arn,
            attributes={'RawMessageDelivery': 'true'}
        )
        mock_prompt.assert_called_once_with(message='Please confirm', skip_prompt=True)

    @patch('montecarlodata.common.resources.prompt_connection')
    def test_create_topic_subscription_with_preexisting_endpoint(self, mock_prompt):
        _SAMPLE_CUST_RESOURCES_CLIENT.list_topic_subscriptions.return_value = [
            [
                {
                    'SubscriptionArn': 'odin',
                    'Owner': '12345',
                    'Protocol': 'sqs',
                    'Endpoint': _SAMPLE_EVENT_DETAILS.event_queue_arn,
                    'TopicArn': _SAMPLE_EVENT_DETAILS.topic_arn
                }
            ]
        ]

        self._service._create_topic_subscription(event_details=_SAMPLE_EVENT_DETAILS, skip_prompts=True)
        _SAMPLE_CUST_RESOURCES_CLIENT.list_topic_subscriptions.assert_called_once_with(
            arn=_SAMPLE_EVENT_DETAILS.topic_arn
        )
        _SAMPLE_CUST_RESOURCES_CLIENT.subscribe_to_topi.cassert_not_called()
        mock_prompt.assert_not_called()

    @patch('montecarlodata.common.resources.AwsClientWrapper')
    def test_get_dc_resource_props(self, aws_wrapper_mock):
        arn = 'arn:aws:cloudformation:us-east-1:1234:stack/foo/bar'
        collector_props = Box({'stackArn': arn})

        self.maxDiff = None
        self.assertEqual(
            self._service.get_dc_resource_props(collector_props=collector_props),
            DcResourceProperties(
                collector_arn=AWSArn(arn),
                collector_props=collector_props,
                collection_region='us-east-1',
                resources_region='us-east-1',
                collection_client=aws_wrapper_mock(),
                resources_client=aws_wrapper_mock()
            )
        )
        aws_wrapper_mock.assert_has_calls(
            [
                call(profile_name=_SAMPLE_CONFIG.aws_profile, region_name=_SAMPLE_CONFIG.aws_region),
                call(profile_name=_SAMPLE_CONFIG.aws_profile, region_name=_SAMPLE_CONFIG.aws_region)
            ]
        )

    def test_parse_stack_prop_list(self):
        self.assertEqual(
            self._service._parse_stack_prop_list(key='foo', val='bar', struct=[{'foo': 'baz', 'bar': 'qux'}]),
            Box({'baz': 'qux'})
        )


import copy
from unittest import TestCase
from unittest.mock import Mock, patch

import click
from box import Box

from montecarlodata.collector.fields import ADD_DC_PROMPT_VERBIAGE
from montecarlodata.collector.management import CollectorManagementService
from montecarlodata.common.data import MonolithResponse
from montecarlodata.common.user import UserService
from montecarlodata.queries.collector import ADD_COLLECTOR_RECORD
from montecarlodata.utils import GqlWrapper, AwsClientWrapper
from tests.helpers import capture_function
from tests.test_common_user import _SAMPLE_CONFIG

_SAMPLE_STACK_ARN = 'stack_arn'
_SAMPLE_GATEWAY_ID = 'gateway_id'
_SAMPLE_TEMPLATE_URI = 'https://s3.amazonaws.com/bucket/public/customer_templates/id.json'
_SAMPLE_INVALID_LAUNCH_URL = f'https://test.com/cf/home?region=us-east-1#/stacks/create/review?stackName=monte-carlo&malformed'
_SAMPLE_VALID_LAUNCH_URL = f'https://test.com/cf/home?region=us-east-1#/stacks/create/review?stackName=monte-carlo&templateURL={_SAMPLE_TEMPLATE_URI}'
_SAMPLE_COLLECTOR_RESPONSE = {
    'generateCollectorTemplate': {
        'dc': {
            'templateLaunchUrl': _SAMPLE_VALID_LAUNCH_URL,
            'stackArn': _SAMPLE_STACK_ARN,
            'active': True,
            'apiGatewayId': _SAMPLE_GATEWAY_ID
        }
    }
}
_SAMPLE_EXTRACTED_COLLECTOR_RESPONSE = _SAMPLE_COLLECTOR_RESPONSE['generateCollectorTemplate']
_SAMPLE_MONOLITH_RESPONSE = MonolithResponse(data=_SAMPLE_EXTRACTED_COLLECTOR_RESPONSE)
_SAMPLE_COLLECTORS_RESPONSE = [
    {
        'uuid': '1234',
        'stackArn': 'aws-loki',
        'active': True,
        'customerAwsAccountId': 'test',
        'templateProvider': 'cloudformation',
        'templateVariant': 'dionysus',
        'templateVersion': '42',
        'codeVersion': '2042',
        'lastUpdated': None
    },
    {
        'uuid': '5678',
        'stackArn': 'aws-thor',
        'active': True,
        'templateProvider': 'terraform',
        'templateVariant': 'foo',
        'templateVersion': '42',
        'codeVersion': '2042',
        'lastUpdated': None
    },
    {
        'uuid': '0912',
        'stackArn': 'aws-odin',
        'active': False,
        'templateProvider': None,
        'templateVariant': None,
        'templateVersion': '42',
        'codeVersion': '2042',
        'lastUpdated': None
    }
]


class CollectorManagementTest(TestCase):
    def setUp(self) -> None:
        self._request_wrapper_mock = Mock(autospec=GqlWrapper)
        self._aws_wrapper_mock = Mock(autospec=AwsClientWrapper)
        self._user_service_mock = Mock(autospec=UserService)

        self._service = CollectorManagementService(
            _SAMPLE_CONFIG,
            request_wrapper=self._request_wrapper_mock,
            aws_wrapper=self._aws_wrapper_mock,
            user_service=self._user_service_mock
        )

    def test_get_template_with_response(self):
        self._request_wrapper_mock.make_request_v2.return_value = None
        with self.assertRaises(click.exceptions.Abort):
            self._service.echo_template()

    def test_get_template_with_invalid_link(self):
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={'dc': {'templateLaunchUrl': _SAMPLE_INVALID_LAUNCH_URL}}
        )

        with self.assertRaises(click.exceptions.Abort):
            self._service.echo_template()

    def test_get_template_with_valid_link(self):
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={'dc': {'templateLaunchUrl': _SAMPLE_VALID_LAUNCH_URL}}
        )

        std_out = capture_function(function=self._service.echo_template).std_out
        self.assertEqual(std_out.getvalue().strip(), _SAMPLE_TEMPLATE_URI)

    def test_upgrade_with_inactive_collector(self):
        sample_collector_response = copy.deepcopy(_SAMPLE_COLLECTOR_RESPONSE)
        sample_collector_response['generateCollectorTemplate']['dc']['active'] = False

        self._request_wrapper_mock.make_request.return_value = sample_collector_response
        with self.assertRaises(click.exceptions.Abort):
            self._service.upgrade_template()

    def test_upgrade_with_missing_fields(self):
        self._request_wrapper_mock.make_request.return_value = {'generateCollectorTemplate': {'dc': {}}}
        with self.assertRaises(click.exceptions.Abort):
            self._service.upgrade_template()

    def test_upgrade(self):
        self._aws_wrapper_mock.get_stack_parameters.return_value = None
        self._aws_wrapper_mock.upgrade_stack.return_value = True

        self._request_wrapper_mock.make_request_v2.return_value = _SAMPLE_MONOLITH_RESPONSE
        self._service.upgrade_template()

        self._aws_wrapper_mock.get_stack_parameters.assert_called_once_with(stack_id=_SAMPLE_STACK_ARN)
        self._aws_wrapper_mock.upgrade_stack.assert_called_once_with(stack_id=_SAMPLE_STACK_ARN,
                                                                     template_link=_SAMPLE_TEMPLATE_URI,
                                                                     parameters=None)
        self._aws_wrapper_mock.deploy_gateway.assert_called_once_with(gateway_id=_SAMPLE_GATEWAY_ID)

    def test_failed_stack_upgrade(self):
        self._aws_wrapper_mock.get_stack_parameters.return_value = None
        self._aws_wrapper_mock.upgrade_stack.return_value = False

        self._request_wrapper_mock.make_request_v2.return_value = _SAMPLE_MONOLITH_RESPONSE
        with self.assertRaises(click.exceptions.Abort):
            self._service.upgrade_template()

        self._aws_wrapper_mock.get_stack_parameters.assert_called_once_with(stack_id=_SAMPLE_STACK_ARN)
        self._aws_wrapper_mock.upgrade_stack.assert_called_once_with(stack_id=_SAMPLE_STACK_ARN,
                                                                     template_link=_SAMPLE_TEMPLATE_URI,
                                                                     parameters=None)
        self._aws_wrapper_mock.deploy_gateway.assert_not_called()

    def test_deploy(self):
        stack_name, termination_protection = 'foo', False
        collector_response = copy.deepcopy(_SAMPLE_COLLECTOR_RESPONSE)
        collector_response['generateCollectorTemplate']['dc']['active'] = False
        monolith_response = MonolithResponse(data=collector_response['generateCollectorTemplate'])

        self._aws_wrapper_mock.create_stack.return_value = True

        self._request_wrapper_mock.make_request_v2.return_value = monolith_response
        self._service.deploy_template(stack_name=stack_name, termination_protection=termination_protection)

        self._aws_wrapper_mock.create_stack.assert_called_once_with(stack_name=stack_name,
                                                                    template_link=_SAMPLE_TEMPLATE_URI,
                                                                    termination_protection=termination_protection,
                                                                    parameters=[])

    def test_deploy_with_an_active_collector(self):
        stack_name, termination_protection = 'foo', True
        self._request_wrapper_mock.make_request_v2.return_value = _SAMPLE_MONOLITH_RESPONSE
        with self.assertRaises(click.exceptions.Abort):
            self._service.deploy_template(stack_name=stack_name, termination_protection=termination_protection)

    def test_replace_params(self):
        existing_params, new_params = [{'ParameterKey': 'foo', 'ParameterValue': '6'}], {'foo': '42'}
        parameters = self._service._build_param_list(existing_params=existing_params, new_params=new_params)
        self.assertEqual(parameters, [{'ParameterKey': 'foo', 'ParameterValue': '42', 'UsePreviousValue': False}])

    def test_only_existing_params(self):
        existing_params = [{'ParameterKey': 'foo', 'ParameterValue': '42'}]
        parameters = self._service._build_param_list(existing_params=existing_params, new_params=None)
        self.assertEqual(parameters, [{'ParameterKey': 'foo', 'UsePreviousValue': True}])

    def test_completely_new_params(self):
        parameters = self._service._build_param_list(existing_params=[], new_params={'foo': '42'})
        self.assertEqual(parameters, [{'ParameterKey': 'foo', 'ParameterValue': '42'}])

    @patch('montecarlodata.collector.management.click')
    def test_get_launch_link(self, click_mock):
        self._request_wrapper_mock.make_request_v2.return_value = _SAMPLE_MONOLITH_RESPONSE
        self._service.launch_quick_create_link(dry=False)

        click_mock.launch.assert_called_once_with(_SAMPLE_VALID_LAUNCH_URL)
        click_mock.echo.assert_not_called()

    @patch('montecarlodata.collector.management.click')
    def test_get_launch_link_dry_run(self, click_mock):
        self._request_wrapper_mock.make_request_v2.return_value = _SAMPLE_MONOLITH_RESPONSE
        self._service.launch_quick_create_link(dry=True)

        click_mock.echo.assert_called_once_with(_SAMPLE_VALID_LAUNCH_URL)
        click_mock.launch.assert_not_called()

    def test_set_region(self):
        region = 'loki'
        service = CollectorManagementService(
            _SAMPLE_CONFIG,
            request_wrapper=self._request_wrapper_mock,
            aws_wrapper=self._aws_wrapper_mock,
            aws_region_override=region
        )
        self.assertEqual(service._collector_region, region)
        self.assertEqual(self._service._collector_region, 'us-east-1')
        self.assertEqual(self._service._collector_region, _SAMPLE_CONFIG.aws_region)

    @patch('montecarlodata.collector.management.click')
    def test_echo_collectors(self, click_mock):
        self._user_service_mock.collectors = _SAMPLE_COLLECTORS_RESPONSE

        self._service.echo_collectors(table_format='plain')
        click_mock.echo.assert_called_once_with(
            'AWS Stack ARN      ID    Version  Template                    Last updated    Active\n'
            'aws-loki         1234       2042  cloudformation:dionysus:42  -               True\n'
            'aws-thor         5678       2042  terraform:foo:42            -               True\n'
            'aws-odin         0912       2042  cloudformation:July-2019    -               False')

    @patch('montecarlodata.collector.management.click')
    def test_echo_collectors_with_only_active(self, click_mock):
        self._user_service_mock.collectors = _SAMPLE_COLLECTORS_RESPONSE

        self._service.echo_collectors(table_format='plain', active_only=True)
        click_mock.echo.assert_called_once_with(
            'AWS Stack ARN      ID    Version  Template                    Last updated    Active\n'
            'aws-loki         1234       2042  cloudformation:dionysus:42  -               True\n'
            'aws-thor         5678       2042  terraform:foo:42            -               True')

    @patch.object(CollectorManagementService, 'launch_quick_create_link')
    @patch('montecarlodata.collector.management.prompt_connection')
    @patch('montecarlodata.collector.management.click')
    def test_add_collector(self, click_mock, prompt_connection_mock, launch_quick_create_link_mock):
        template_provider = 'provider'
        template_variant = 'variant'
        mocked_region = 'us-east-1'
        click_mock.prompt.return_value = mocked_region
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(data=Box({'dc': {'uuid': '1234'}}))

        self._service.add_collector(
            template_provider=template_provider,
            template_variant=template_variant
        )

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=ADD_COLLECTOR_RECORD,
            operation='createCollectorRecord',
            variables = {
                'template_provider': template_provider,
                'template_variant': template_variant
            }
        )
        click_mock.echo.assert_called_once_with("Created collector record with ID '1234'")
        prompt_connection_mock.assert_called_once_with(message=ADD_DC_PROMPT_VERBIAGE)
        launch_quick_create_link_mock.assert_called_once_with(
            dry=False,
            dc_id='1234',
            collection_region=mocked_region
        )

    @patch('montecarlodata.collector.management.click')
    def test_add_collector_no_prompt(self, click_mock):
        template_provider = 'provider'
        template_variant = 'variant'

        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(data=Box({'dc': {'uuid': '1234'}}))

        self._service.add_collector(template_provider=template_provider,
                                    template_variant=template_variant,
                                    no_prompt=True)

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=ADD_COLLECTOR_RECORD,
            operation='createCollectorRecord',
            variables = {
                'template_provider': template_provider,
                'template_variant': template_variant
            }
        )
        click_mock.echo.assert_called_once_with("Created collector record with ID '1234'")
